#include "fsyslogd.h"

int Nhash;
struct conf_t conf;
struct ip_t *ips, *lookup[MAXHASH];
unsigned long int drop;

void ReloadConfig() {
	CloseFiles();
	OpenFiles();
	signal( SIGHUP, (void *)ReloadConfig );
}

static void log(int priority, const char *format, va_list args) {
char buffer[MAXBUFFER];

	vsnprintf(buffer, sizeof(buffer), format, args);

	if( conf.daemon )
		syslog(priority, "%s", buffer);
	else
		printf( "%s\n", buffer );
}
	
void LogInfo(const char *format, ...) {
va_list args;
	
	va_start(args, format);
	log(LOG_INFO, format, args);
	va_end(args);
}
	
void LogError(const char *format, ...) {
va_list args;
	
	va_start(args, format);
	log(LOG_ERR, format, args);
	va_end(args);
}

void PrintStat() {
	LogInfo( "Packet dropped: %l", drop );
	signal( SIGUSR1, (void *)PrintStat );
}

void End() {
	CloseFiles();
	LogInfo( "Game over" );

	if( conf.daemon ) closelog();

exit(0);	
}

/*************************************************************************************
* Close files to write
***************************************************************************************/
int CloseFiles() {
struct ip_t *p_ips;

   for( p_ips = ips; p_ips; p_ips = p_ips->next ) {
      if( close(FDS(p_ips->addr_ip,Nhash)) < 0 )
         {
            LogInfo( "Warning: Can't close file '%s'", p_ips->filename );
            continue;
         }
   }

return(0);
}

/*************************************************************************************
* Open files to write
***************************************************************************************/
int OpenFiles() {
struct ip_t *p_ips;

   for( p_ips = ips; p_ips; p_ips = p_ips->next ) {
      if( (FDS(p_ips->addr_ip,Nhash) = open( p_ips->filename,
            O_WRONLY|O_CREAT|O_NOCTTY|O_NONBLOCK, S_IWRITE|S_IREAD|S_IRGRP)) == -1 )
         {
            LogInfo( "Warning: Can't open file '%s'", p_ips->filename );
            continue;
         }
         lseek( FDS(p_ips->addr_ip,Nhash), 0, SEEK_END );
   }

return(0);
}

/*************************************************************************************
* Load IPs to syslogd
***************************************************************************************/
int LoadIPs() {
int i = 0;
struct ip_t *p_ips, *before = NULL;
char buffer[MAXBUFFER+1], *ip, *filename;
FILE *fd;

	if( !(fd = fopen(conf.config_file?conf.config_file:"fsyslogd.conf", "r") ) ) {
		printf( "Error opening config file\n" );
		exit(-1);
	}

	while( fgets( buffer, sizeof(buffer), fd) ) {
		if( !strlen(buffer) || ( buffer[strlen(buffer)-1] != '\n' && feof(fd) ) ) {
			printf( "Error reading config file\n" );
			exit(-1);
		}

		if( buffer[0] == '#' ) continue;

		if( !(ip = strtok(buffer, " \n\t")) ) {
				printf( "Error parsing config file (ip)\n" );
				exit(-1);
		}

		if( !(filename = strtok(NULL, " \n\t"))) {
			printf( "Error parsing config file (filename)\n" );
			exit(-1);
		}

		p_ips = malloc( sizeof(struct ip_t) );
		p_ips->addr_ip = inet_addr(ip);
		p_ips->string_ip = malloc( strlen(ip) );
		p_ips->filename = malloc( strlen(filename) );
		p_ips->next = NULL;

		strcpy( p_ips->string_ip, ip );
		strcpy( p_ips->filename, filename );

		if( ips == NULL ) ips = p_ips;
		else before->next = p_ips;

		before = p_ips;

		i++;
	}

return( i );
}

/*************************************************************************************
* Calculate number N to use in Hash function
***************************************************************************************/
int CalculaNhash(int maxips) {
int i, n;
struct ip_t *p_ips;

	for( p_ips = ips, n = MAXHASH; n > maxips && p_ips; n-- ) {

		for( i = 0; i < MAXHASH; i++) lookup[i] = NULL;

		for( p_ips = ips; p_ips; p_ips = p_ips->next ) {

			if( LOOKUP(p_ips->addr_ip, n) ) break;

			LOOKUP(p_ips->addr_ip, n) = p_ips;
		}

	}

	if( p_ips ) return -1;

return(++n);
}

int get_gid(char *group) {
struct group *gr;
        
	if (group) {
		gr = getgrnam(group);
		if (gr) return( gr->gr_gid );
			else return( -1 );
	} else return( -1 );

return( -1 );
}

int get_uid(char *user) {
struct passwd *pw;
        
	if (user) {
		pw = getpwnam(user);
		if(pw) return(pw->pw_uid);
			else return(-1);
	} else return(-1);
}

int InitSyslog() {
struct sockaddr_in servAddr;
int s, uid, gid;

	if((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
		LogError( "Cannot open socket\n" );
		exit(-1);
	}

	servAddr.sin_family = AF_INET;
	servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servAddr.sin_port = htons(conf.port);

	if( bind(s, (struct sockaddr *)&servAddr, sizeof(servAddr)) < 0 ) {
		LogError("Cannot bind port number %d\n", conf.port);
		exit(-1);
  }

	if( conf.user ) {
		if( (uid = get_uid(conf.user)) < 0 ) {
			LogError( "Error: User %s doesn't exist\n", conf.user );
			exit(-1);
		}
		setuid( uid );
	}

	if( conf.group ) {
		if ((gid = get_gid(conf.group)) < 0 ) {
			LogError( "Error: Group %s doesn't exist\n", conf.group );
			exit(-1);
		}
		setgid( gid );
	}

	if( conf.chroot ) {
		chdir( conf.chroot ); 
		chroot( "." ); 
	}

return(s);
}

int RunSyslog( int s ) {
time_t now;
register int n, cliLen;
struct ip_t *ip;
struct sockaddr_in cliAddr;
char msg[MAXMSG+1];

	cliLen = sizeof(cliAddr);

	/* receive message */
	while( 1 ) {
		if( (n = recvfrom(s, msg, MAXMSG, 0, (struct sockaddr *) &cliAddr, &cliLen)) < 0 ) {
			LogInfo( "Error receiving packet\n" );
			continue;
		}

		if( n > 1024 ) {
			drop++;
			continue;
		}

		if( !(ip = lookup[HASH(cliAddr.sin_addr.s_addr, Nhash)]) ) continue;

		if( ip->addr_ip != cliAddr.sin_addr.s_addr ) continue;

		time( &now );
		write( ip->fd, ctime(&now), 20 );
		write( ip->fd, msg, n );
	}

return(0);
}

void usage(char *progname) {

	fprintf(stderr, "%s version "VERSION" by Fryxar\n\n", progname);
	fprintf(stderr, "Usage: %s [-hdf] [-p port] [-u user] [-g group] [-r chroot_dir] [-c config_file]\n", progname);
	fprintf(stderr, "          -h               : This help\n" );
	fprintf(stderr, "          -d               : Debug mode\n" );
	fprintf(stderr, "          -f               : Don't fork into background\n" );
	fprintf(stderr, "          -p <pnumber>     : Set port to listen\n" );
	fprintf(stderr, "          -c <fname>       : Set config file name\n" );
	fprintf(stderr, "          -r <dir>         : Chroot to directory\n" );
	fprintf(stderr, "          -g <group>       : Switch to user\n" );
	fprintf(stderr, "          -u <user>        : Switch to group\n" );

exit(0);
}

void InitDaemon( char *progname, int pid ) {

   if( pid ) exit( 0 );

   umask( 026 );

   if( setsid() == -1 ) {
       printf( "Can't setsid\n" );
       exit( -1 );
   }

	openlog( progname, LOG_PID, LOG_USER );

   /* close file descriptors */
   close( STDIN_FILENO );                        /* STDIN */
   close( STDOUT_FILENO );                       /* STDERR */
   close( STDERR_FILENO );                       /* STDERR */

   open( "/dev/null", O_RDWR );                  /* STDIN */
   dup( STDIN_FILENO );                          /* STDOUT */
   dup( STDIN_FILENO );                          /* STDERR */
}

/*************************************************************************************
* MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN 
***************************************************************************************/
int main(int argc, char *argv[]) {
int maxips;
int arg, pid, s;

	conf.debug = 0;
	conf.daemon = 1;
	conf.port = DEFAULT_PORT;

	opterr = 0;
	while ((arg = getopt(argc, argv, "fhdc:u:g:r:")) != EOF) {
		switch (arg) {
			case 'h':
				usage(argv[0]);
				break;
			case 'c':
				conf.config_file = optarg;
				break;
			case 'r':
				conf.chroot = optarg;
				break;
			case 'u':
				conf.user = optarg;
				break;
			case 'p':
				conf.port = atoi(optarg);
				break;
			case 'g':
				conf.group = optarg;
				break;
			case 'f':
				conf.daemon = 0;
				break;
			case 'd':
				conf.debug = 0;
				break;
			default:
				printf("error: unrecognized switch '%c'", optopt);
				usage(argv[0]);
				break;
		}
	}

	maxips = LoadIPs();

	if( (Nhash = CalculaNhash(maxips)) == -1 ) {
		printf( "Can't use hash function\n" );
		exit(-1);
	}	

	/* Start Daemon */
	if( conf.daemon ) {
		if( ( pid = fork() ) < 0 ) {
			printf( "Error when forking\n" );
			exit(-1);
		}

		InitDaemon( argv[0], pid );
	}

	signal( SIGHUP, (void *)ReloadConfig );
	signal( SIGUSR1, (void *)PrintStat );
	signal( SIGTERM, (void *)End );
	signal( SIGINT, (void *)End );

	s = InitSyslog();

	OpenFiles();

	RunSyslog( s );

	CloseFiles();

return 0;
}
